/* Gnomba
 * Copyright (C) 1999 Gnomba Team
 *
 * net.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include "nmbhdr.h"
#include "net.h"
#include "error.h"
#include "smbwrap.h"

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define BrokenLookup	false
#define SlowScan	false
#define debug		false
static const char *DefaultWorkgroup = NULL;


#define initSocks()
#define closeSocks()
typedef int SOCKET;
#define MAXSOCKETS 64

machineNode::machineNode()
{
	machineName = NULL;
	workgroupName = NULL;
	ip = NULL;
	next = NULL;
}
machineNode::~machineNode()
{
	free (machineName);
	free (workgroupName);
	free (ip);
}

void trim(char * str)
{

  int i;

  if (str == NULL)
    return;
  for (i = strlen(str) - 1; i >= 0; i--)
    if (str[i] != ' ' && str[i] != '\t')
      break;
  str[i + 1] = '\0';
}


void sock_unblock(SOCKET s)
{
#ifdef WIN32
  u_long flag = 1;		/* (doblock ? 0 : 1) */
  ioctlsocket(s, FIONBIO, &flag);
#else
  int flag = O_NONBLOCK | fcntl(s, F_GETFL);
  fcntl(s, F_SETFL, flag);
#endif
}

void sock_block(SOCKET s)
{
#ifdef WIN32
  u_long flag = 0;		/* (doblock ? 0 : 1) */
  ioctlsocket(s, FIONBIO, &flag);
#else
  int flag = (~O_NONBLOCK) & fcntl(s, F_GETFL);
  fcntl(s, F_SETFL, flag);
#endif
}

SOCKET add_sock(u_long ip, u_short port)
{
	struct sockaddr_in dest;
	struct linger slinger;

	memset(&dest, 0, sizeof(struct sockaddr_in));

	SOCKET sfd = socket(AF_INET, SOCK_STREAM, 0);
	while (sfd < 0) {
		usleep(10000);
		sfd = socket(AF_INET, SOCK_STREAM, 0);
	}

	slinger.l_onoff = 1;
	slinger.l_linger = 0;
	setsockopt(sfd, SOL_SOCKET, SO_LINGER, (char *) &slinger, sizeof(slinger));
	dest.sin_family = AF_INET;
	dest.sin_port = htons(port);
	dest.sin_addr.s_addr = htonl(ip);
	sock_unblock(sfd);
	if (connect(sfd, (struct sockaddr *) &dest, sizeof(struct sockaddr_in)) < 0) {
		if ((errno == EWOULDBLOCK) || (errno == EINPROGRESS)) {
			sock_block(sfd);
			return sfd;
		}
		close (sfd);
		if (errno == ECONNREFUSED) return 0;
		return 0;
	}
	sock_block(sfd);
	return sfd;
}


int getNMBInfo(u_long ip, char *machine, char *workgroup)
{

  struct sockaddr_in sin_dst, sin_src;
  struct nmbhdr *nmb;
  char *data;
  struct typez *typz;
  char buffer[1024];
  int socket_client, i, timeout = 0;
  socklen_t longueur = sizeof(struct sockaddr_in);
  int count;
	int loops;
	int optval;
  struct _nameinfo {
    char name[15];
    char type;
    char res;
    char fill;
  } *ninf;


  bzero(buffer, sizeof(buffer));
  nmb = (struct nmbhdr *) buffer;
  data = (char *) (buffer + NMBHDRSIZE);
  typz = (struct typez *) (buffer + NMBHDRSIZE + 33);

  sin_dst.sin_family = AF_INET;
  sin_dst.sin_port = htons(137);
  sin_dst.sin_addr.s_addr = ip;

  memcpy(data, "CKAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", 34);

  nmb->id = 0x600;
  nmb->R = 0;			/* 0 for question 1 for response */
  nmb->opcode = 0;		/* 0 = query */
  nmb->que_num = htons(1);	/* i have only 1 question :) */
  nmb->namelen = 0x20;
  typz->type = 0x2100;
  typz->type2 = 0x0100;

  socket_client = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

  if (BrokenLookup) {
	  optval=1;
		setsockopt(socket_client,SOL_SOCKET,SO_REUSEADDR,&optval,sizeof(optval));
		sin_src.sin_port = htons (137);
		sin_src.sin_addr.s_addr = INADDR_ANY;
		if (bind(socket_client,(struct sockaddr *) &sin_src,sizeof(sin_src))==(-1))
		  {
	      perror("bind");
			  return(-1);
			}
	}
	
	
  sendto(socket_client, buffer, 50, 0, (struct sockaddr *) &sin_dst, longueur);
  usleep(100000);
  sendto(socket_client, buffer, 50, 0, (struct sockaddr *) &sin_dst, longueur);

  bzero(buffer, sizeof(buffer));
  nmb = (struct nmbhdr *) (buffer);
  machine[0] = '\0';
  workgroup[0] = '\0';

  sock_unblock(socket_client);

	if (SlowScan) loops = 50; else loops = 2;
  for (timeout = 0; timeout < loops; timeout++) {
    if (SlowScan){
	usleep(700000);
    }else{
	usleep(100000);
    }
    if (recvfrom(socket_client, buffer, sizeof(buffer), 0, (struct sockaddr *) &sin_dst, &longueur) != -1) {
      if (nmb->rep_num != 0) {

	/*set to first nameblock after header */
	count = 0;
	ninf = (struct _nameinfo *) (buffer + 57 + (count++ * sizeof(struct _nameinfo)));
	while ((machine[0] == '\0') && (57 + (count * sizeof(struct _nameinfo)) < 1024)) {
	  if (ninf->type == 0) {
	    i = 0;
	    while ((i < 15)) {
	      machine[i] = ninf->name[i];
	      i++;
	    }
	    machine[i] = '\0';
	    trim(machine);
	  }
	  ninf = (struct _nameinfo *) (buffer + 57 + (count++ * sizeof(struct _nameinfo)));
	}
	if (machine[0] == '\0') {
	  if (debug)
	    printf("we finished without finding anything, giving up\n");
	  close(socket_client);
	  return (-1);
	}
	while ((workgroup[0] == '\0') && (57 + (count * sizeof(struct _nameinfo)) < 1024)) {
	  if (ninf->type == 0) {
	    i = 0;
	    while ((i < 15)) {
	      workgroup[i] = ninf->name[i];
	      i++;
	    }
	    workgroup[i] = '\0';
	    trim(workgroup);
	  }
	  ninf = (struct _nameinfo *) (buffer + 57 + (count++ * sizeof(struct _nameinfo)));
	}
	if (workgroup[0] == '\0') {
	  if (debug)
	    printf("we finished without finding anything, giving up\n");
	  close(socket_client);
	  return (-1);
	}
	close(socket_client);
	return 0;
      }
      close(socket_client);
      return (-1);
    }
  }
  close(socket_client);
  return (-1);
}


machineNode *gather(
	SOCKET * s,
	SOCKET max,
	int cnt,
	u_long ipstart,
	fd_set ofds,
	struct timeval * tv,
	int retries,
	machineNode * root)
{
	if (SlowScan) usleep(100000); else usleep(50000);

	long timeout = tv->tv_sec + (tv->tv_usec * 100000);
	struct timeval begin;
	gettimeofday(&begin, NULL);
	int sel = 1;

	while ((sel != 0) || (retries) || timeout) {
		fd_set wfds = ofds;
		int res = select(max + 1, NULL, &wfds, NULL, tv);
		sel = res;
		if (!res) --retries;
		for (int i = 0; res && i < cnt; ++i)
		if ((res > 0) && FD_ISSET(s[i], &wfds)) {
			u_char buf[10];
			if (send(s[i], buf, 0, 0) == 0) {
				u_long currip = ipstart + i;
				u_char *ip = (u_char *) & currip;
				char nbt_name[25];
				char grp_name[25];
				if (getNMBInfo(htonl(currip), nbt_name, grp_name) >= 0) {
					if (debug){
						printf("%d.%d.%d.%d --> %s\\\\%s\n", ip[3], ip[2], ip[1], ip[0], grp_name, nbt_name);
					}

					if ( ((DefaultWorkgroup && *DefaultWorkgroup) &&
						!strncmp( grp_name, DefaultWorkgroup, strlen(grp_name))) ||
						!DefaultWorkgroup ||
						(DefaultWorkgroup && !*DefaultWorkgroup))
						{
						machineNode *curr = new machineNode;
						curr->next = root;

						curr->ip = (char*)malloc(sizeof(char) * 16);
						sprintf(curr->ip, "%d.%d.%d.%d", ip[3], ip[2], ip[1], ip[0]);

						curr->machineName = strdup(nbt_name);
						curr->workgroupName = strdup (grp_name);
						curr->scanned = false;

						root = curr;
					}else{
						if (debug){
							printf("%d.%d.%d.%d <NoName>\n", ip[3], ip[2], ip[1], ip[0]);
						}
					}
				}
			}
			--res;
			FD_CLR(s[i], &ofds);
			close(s[i]);
		}
		#if 0
			struct timeval end;
			gettimeofday(&end, NULL);
			long diff = (end.tv_sec + (100000 * end.tv_usec)) - (begin.tv_sec + (100000 * begin.tv_usec));
		#endif
		timeout -= timeout;
	}
	return root;
}

machineNode *scanloop(const char *strrange, long delay, machineNode * root)
{
	machineNode *curr = root;
	initSocks();

	signal(SIGPIPE, SIG_IGN);

	int input[8];
	sscanf(strrange, "%d.%d.%d.%d %d.%d.%d.%d", (int *) &input[0], (int *) &input[1], (int *) &input[2], (int *) &input[3], (int *) &input[4], (int *) &input[5], (int *) &input[6], (int *) &input[7]);

	if (debug){
		fprintf(stderr,"IP: %d.%d.%d.%d - %d.%d.%d.%d\n", input[0], input[1], input[2], input[3], input[4], input[5], input[6], input[7]);
	}

	u_long ipstart = (input[0] << 24) + (input[1] << 16) + (input[2] << 8) + input[3];
	u_long ipstop = (input[4] << 24) + (input[5] << 16) + (input[6] << 8) + input[7];
	struct timeval tv = {0, 0};
	tv.tv_usec = delay;
	int range = ipstop - ipstart + 1;
	if (range <= 0) return curr;
	SOCKET *s = (SOCKET *) malloc(sizeof(SOCKET) * ((range < MAXSOCKETS) ? range : MAXSOCKETS));

	if (s == NULL) {
		printf("Unable to alloc memory.\n");
		exit(-1);
	}
	bzero(s, sizeof(SOCKET) * ((range < MAXSOCKETS) ? range : MAXSOCKETS));

	do {
		fd_set origfds;
		FD_ZERO(&origfds);
		u_short i = 0;
		SOCKET max = 0;
		for (; (i < range) && (i < MAXSOCKETS); ++i) {
			s[i] = add_sock(ipstart + i, 139);
			if (s[i] < 1) {
				continue;
			}
			max = (s[i] > max) ? s[i] : max;
			if (s) FD_SET(s[i], &origfds);
		}
		curr = gather(s, max, i, ipstart, origfds, &tv, 5, curr);

		range -= i;
		ipstart += i;
		for (; i > 0; --i) close(s[i]);
		close(s[0]);
	} while (range);
	closeSocks();
	free(s);
	return curr;
}


machineNode * netscan(
	const char *range[],
	int nbrange)
{
	machineNode *root = NULL;
  
	if ((BrokenLookup)&&(getuid()!=0)&&(geteuid()!=0)) { /*Test this now so we only show one warning*/
		fprintf (stderr,"Must be root to bind to ports below 1024\n");
		return NULL;
	}
 
	for (int row = 0; row < nbrange; row++) {
		const char *text = range[row];
		if (SlowScan){
			root = scanloop(text, 20000, root);
		}else{
			root = scanloop(text, 1000, root);
		}
	}
	return root;
}

/*
	Delete all the link list
*/
void net_delnode (machineNode *root)
{
	while (root != NULL){
		machineNode *next = root->next;
		delete root;
		root = next;
	}
}

#if 0
int main(int argc, char *argv[])
{
	machineNode *node = netscan ((const char **)argv+1,argc-1);
	while (node != NULL){
		printf ("Node %s %s\n",node->machineName,node->workgroupName);
		node = node->next;
	}
	return 0;
}
#endif
