#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <pwd.h>
#include <stdlib.h>

static char *str_skip (char *buf)
{
	while (isspace(*buf)) buf++;
	return buf;
}
static char *str_copyword (char *dst, char *pt)
{
	while (*pt > ' ') *dst++ = *pt++;
	*dst = '\0';
	return pt;
}

static char *strip_end (char *buf)
{
	int last = strlen(buf)-1;
	while (last >= 0 && isspace(buf[last])) buf[last--] = '\0';
	return last == 0 ? buf : buf + last + 1;
}

/*
	Read a line of a configuration file and process continuation line.
	Return buf, or NULL si eof
	Blank at the end of line are always stripped.
	Everything following comcar is a comment. The line is cut before.
*/
static char *fgets_strip (
	char *buf,
	int sizebuf,
	FILE *fin,
	char contcar,		// Continuation char, generally backslash
	char comcar,		// Comment character, generally #
	int *noline,		// This will be updated and contain the line number
						// Can be NULL
	int *empty)			// Will be != 0 if the line is empty and did not
						// even hold a comment.
{
	int nocomment = 1;		// No comments found ?
	int contline=0;
	char *debut = buf;
	char *ret = NULL;
	*buf = '\0';
	*empty = 1;
	while (fgets(buf,sizebuf,fin)!=NULL){
		char *end = strip_end (buf);
		char *pt = strchr(buf,comcar);
		if (pt != NULL){
			nocomment = 0;
			*pt = '\0';
			end = strip_end (buf);
		}
		if (noline != NULL) (*noline)++;
		ret = debut;
		if (contline){
			char *pt = str_skip(buf);
			if (pt > buf+1){
				strcpy (buf+1,pt);
				buf[0] = ' ';
				end-=(int)(pt-buf)-1;
			}else if (pt == buf+1){
				buf[0] = ' ';
			}
		}
		if (end > buf && *(end-1) == contcar){
			if (end == buf+1 || *(end-2) != contcar){
				/* Continuation demand */
				contline = 1;
				end--;
				*end = '\0';
				buf = end;
			}else{
				*(end-1) = '\0';
				break;
			}
		}else{
			break;
		}
	}
	*empty = debut[0] == '\0' && nocomment;
	return ret;
}		



int main (int argc, char *argv[])
{
	int ret = -1;
	if ((argc == 2 || argc == 3) && strcmp(argv[1],"--list")==0){
		if (argc == 2){
			FILE *fin = fopen ("/etc/printcap","r");
			if (fin != NULL){
				char buf[2000];
				int empty;	
				while (fgets_strip(buf,sizeof(buf)-1,fin,'\\','#',NULL,&empty)!=NULL){
					char *pt = strchr(buf,'|');
					if (pt != NULL){
						*pt = '\0';
					}else{
						pt = strchr(buf,':');
						if (pt != NULL) *pt = '\0';
					}
					pt = str_skip(buf);
					if (*pt != '\0') printf ("d %s %d %d 0\n",pt,getuid(),getgid());
				}
				fclose (fin);
				ret = 0;
			}
		}else if (argc == 3){
			char cmd[strlen(argv[1])+10];
			snprintf (cmd,sizeof(cmd)-1,"lpq -P%s",argv[2]);
			FILE *fin = popen (cmd,"r");
			if (fin != NULL){
				char buf[10000];
				bool header = true;
				while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
					if (header){
						if (strncmp(buf,"Rank",4)==0){
							header = false;
						}
					}else{
						char *pt = buf;
						while (*pt > ' ') pt++;
						pt = str_skip(pt);
						char owner[200],job[200];
						pt = str_copyword (owner,pt);
						pt = str_skip(pt);
						pt = str_copyword (job,pt);
						pt = str_skip(pt);
						int uid = 0;
						struct passwd *p = getpwnam(owner);
						if (p != NULL) uid = p->pw_uid;
						int size = 0;
						pt = strstr (buf,"bytes");
						if (pt != NULL){
							while (isspace(pt[-1])) pt--;
							while (isdigit(pt[-1])) pt--;
							size = atoi(pt);
						}
						printf ("f %s %d 0 %d\n",job,uid,size);
					}
				}
				pclose (fin);
				ret = 0;
			}
		}
	}else if (argc == 3 && strcmp(argv[1],"--unlink")==0){
		int len = strlen(argv[2])+1;
		char buf[len];
		strcpy (buf,argv[2]);
		char *pt = strchr(buf,'/');
		if (pt == NULL){
			fprintf (stderr,"unlink: invalid printer/job path %s\n",buf);
		}else{
			*pt++ = '\0';
			char cmd[len+10];
			snprintf (cmd,sizeof(cmd)-1,"lprm -P%s %s",buf,pt);
			ret = system (cmd);
		}
	}
	return ret;
}

